import logging
import re
from decimal import Decimal
from queue import Empty, Queue
from threading import Event as ThreadEvent
from typing import Optional

from .event import SilenceEvent
from .stream import StreamListenerHandler

logger = logging.getLogger(__name__)

SILENCE_DETECT_RE = re.compile(
    r"\[silencedetect.*\] silence_(start|end): (\d+(?:\.\d+)?)(?: \| silence_duration: (\d+\.\d+))?"
)

# https://ffmpeg.org/ffmpeg-filters.html#silencedetect
DEFAULT_SILENCE_DETECT_RAW: str = "silencedetect=noise={noise}:d={duration}"
DEFAULT_SILENCE_DETECT_NOISE: str = "-60dB"
# https://ffmpeg.org/ffmpeg-utils.html#time-duration-syntax
DEFAULT_SILENCE_DETECT_DURATION: str = "2.0"


def parse_silence_detect(line: str) -> Optional[SilenceEvent]:
    line = line.strip()
    match = SILENCE_DETECT_RE.search(line)
    if match is None:
        return None

    logger.debug(line)

    return SilenceEvent(
        kind=match.group(1),  # type: ignore[arg-type]
        seconds=Decimal(match.group(2)),
        duration=Decimal(match.group(3)) if match.group(3) else None,
    )


# pylint: disable=too-few-public-methods
class SilenceHandler(StreamListenerHandler):
    name = "silence_handler"
    stop: ThreadEvent
    queue: Queue

    event_queue: Queue

    noise: str
    duration: str
    raw: str

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        stop: ThreadEvent,
        event_queue: Queue,
        noise: str = DEFAULT_SILENCE_DETECT_NOISE,
        duration: str = DEFAULT_SILENCE_DETECT_DURATION,
        raw: str = DEFAULT_SILENCE_DETECT_RAW,
    ):
        super().__init__()
        self.stop = stop
        self.queue = Queue()

        self.event_queue = event_queue
        self.noise = noise
        self.duration = duration
        self.raw = raw

    def ffmpeg_output(self):
        return [
            *("-af", self.raw.format(noise=self.noise, duration=self.duration)),
            *("-f", "null", "/dev/null"),
        ]

    def run(self):
        logger.info("starting %s", self.name)

        # pylint: disable=duplicate-code
        while not self.stop.is_set() or not self.queue.empty():
            try:
                line = self.queue.get(timeout=1.0)
            except Empty:
                continue

            event = parse_silence_detect(line)
            if event is not None:
                logger.info(f"silence {event.kind}: {event.when}")
                self.event_queue.put(event)

            self.queue.task_done()

        logger.info("%s stopped", self.name)



import logging
import re
from itertools import chain
from pathlib import Path
from queue import Empty, Queue
from shutil import move
from threading import Event as ThreadEvent
from typing import Optional

from typing_extensions import Protocol

from .prometheus import archive_errors, archive_segments
from .stream import StreamListenerHandler

logger = logging.getLogger(__name__)

DEFAULT_ARCHIVE_SEGMENT_SIZE: int = 3600
DEFAULT_ARCHIVE_SEGMENT_FORMAT: str = "ogg"

DEFAULT_ARCHIVE_SEGMENT_FILEPATH: str = (
    "{year}/{month}/{day}/{hour}{minute}{second}.{format}"
)

SEGMENT_STARTED_RE = re.compile(
    r"\[segment @ .*\] segment:'(.*)' starts with packet .*",
)
SEGMENT_ENDED_RE = re.compile(
    r"\[segment @ .*\] segment:'(.*)' count:\d+ ended",
)


def parse_segment_line(regex: re.Pattern, line: str) -> Optional[Path]:
    line = line.strip()
    match = regex.search(line)
    if match is None:
        return None

    return Path(match.group(1))


class IngestSegmentError(Exception):
    """An error occurred while ingesting a segment"""


class InvalidSegmentFilename(IngestSegmentError):
    """Invalid segment filename"""


# pylint: disable=too-few-public-methods
class ArchiveStorage(Protocol):
    def ingest_segment(self, segment: Path, segment_filepath: Path): ...


# pylint: disable=too-few-public-methods
class LocalArchiveStorage:
    path: Path

    def __init__(self, path: str) -> None:
        self.path = Path(path)
        if not self.path.is_dir():
            raise ValueError(f"archive path '{path}' is not a directory!")

    def ingest_segment(self, segment: Path, segment_filepath: Path):
        segment_fullpath = self.path / segment_filepath

        try:
            logger.debug(f"moving segment to {segment_fullpath}")
            segment_fullpath.parent.mkdir(parents=True, exist_ok=True)
            move(segment, segment_fullpath)
        except OSError as exception:
            raise IngestSegmentError(exception) from exception


# pylint: disable=too-many-instance-attributes
class ArchiveHandler(StreamListenerHandler):
    name = "archive_handler"
    stop: ThreadEvent
    queue: Queue

    SEGMENTS_DIR: str = "incoming"
    SEGMENTS_PENDING_DIR: str = "pending"
    SEGMENT_PREFIX: str = "segment"
    SEGMENT_STRFTIME: str = "%Y-%m-%d-%H-%M-%S"

    storage: ArchiveStorage

    segment_filepath: str
    segment_size: int
    segment_format: str
    segment_format_options: Optional[str]
    copy_stream: bool

    segments_pending_queue: set[Path]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        stop: ThreadEvent,
        storage: ArchiveStorage,
        segment_filepath: str = DEFAULT_ARCHIVE_SEGMENT_FILEPATH,
        segment_size: int = DEFAULT_ARCHIVE_SEGMENT_SIZE,
        segment_format: str = DEFAULT_ARCHIVE_SEGMENT_FORMAT,
        segment_format_options: Optional[str] = None,
        copy_stream: bool = False,
    ):
        super().__init__()
        self.stop = stop
        self.queue = Queue()

        self.storage = storage

        self.segment_filepath = segment_filepath
        self.segment_size = segment_size
        self.segment_format = segment_format
        self.segment_format_options = segment_format_options
        self.copy_stream = copy_stream

        self.segments_dir = Path(self.SEGMENTS_DIR)
        self.segments_pending_dir = Path(self.SEGMENTS_PENDING_DIR)

        self.segments_dir.mkdir(exist_ok=True)
        self.segments_pending_dir.mkdir(exist_ok=True)

        self.segments_pending_queue = set()

    def _tmp_segment_filename(self) -> Path:
        return Path(
            ".".join(
                [
                    self.SEGMENT_PREFIX,
                    self.SEGMENT_STRFTIME,
                    self.segment_format,
                ]
            )
        )

    def _segment_filepath(self, segment: Path) -> Path:
        try:
            filename_parts = segment.name.split(".")
            datetime_str = filename_parts[1]
            datetime_parts = datetime_str.split("-")

            return Path(
                self.segment_filepath.format(
                    **{
                        "format": self.segment_format,
                        "year": datetime_parts[0],
                        "month": datetime_parts[1],
                        "day": datetime_parts[2],
                        "hour": datetime_parts[3],
                        "minute": datetime_parts[4],
                        "second": datetime_parts[5],
                    }
                )
            )
        except IndexError as error:
            raise InvalidSegmentFilename(
                f"found invalid segment filename {segment.name}"
            ) from error

    def ffmpeg_output(self):
        args = ["-map_metadata", "-1"]

        if self.copy_stream:
            logger.info("enabling stream copy")
            args += ("-c", "copy")

        args += (
            *("-f", "segment"),
            *("-strftime", "1"),
            *("-segment_time", str(self.segment_size)),
            *("-segment_format", self.segment_format),
        )

        if self.segment_format_options:
            args += ("-segment_format_options", self.segment_format_options)

        args += (
            *("-reset_timestamps", "1"),
            str(self.segments_dir / self._tmp_segment_filename()),
        )

        return args

    def gather_unmanaged_segments(self):
        for segment_incoming in chain(
            self.segments_dir.iterdir(),
            self.segments_pending_dir.iterdir(),
        ):
            self.accept_incoming_segment(segment_incoming)

    def accept_incoming_segment(self, segment_incoming: Path) -> Path:
        segment_pending = self.segments_pending_dir / segment_incoming.name

        if segment_incoming != segment_pending:
            logger.debug(f"moving {segment_incoming} to {segment_pending}")
            move(segment_incoming, segment_pending)

        self.segments_pending_queue.add(segment_pending)

        return segment_pending

    def _ingest_segment(self, segment: Path):
        with archive_errors.count_exceptions(IngestSegmentError):
            self.storage.ingest_segment(
                segment,
                self._segment_filepath(segment),
            )
            archive_segments.labels(state="ingested").inc()

    def ingest_pending_segments(self):
        while self.segments_pending_queue:
            segment = self.segments_pending_queue.pop()

            try:
                self._ingest_segment(segment)
            except IngestSegmentError as exception:
                self.segments_pending_queue.add(segment)

                logger.error(exception)

                # Go try to ingest segments from incoming, pending can wait
                break

    def run(self):
        logger.info("starting %s", self.name)

        # Ingest pending segments before starting
        self.gather_unmanaged_segments()
        self.ingest_pending_segments()

        while not self.stop.is_set() or not self.queue.empty():
            try:
                line = self.queue.get(timeout=1.0)
            except Empty:
                continue

            segment_started = parse_segment_line(SEGMENT_STARTED_RE, line)
            if segment_started is not None:
                logger.info("segment started %s", segment_started)
                self.queue.task_done()
                continue

            segment_ended = parse_segment_line(SEGMENT_ENDED_RE, line)
            if segment_ended is None:
                self.queue.task_done()
                continue

            logger.info("segment ended %s", segment_ended)

            try:
                self.accept_incoming_segment(segment_ended)
                archive_segments.labels(state="incoming").inc()
                self.queue.task_done()

                self.ingest_pending_segments()
            except InvalidSegmentFilename as exception:
                logger.error(exception)
                continue
            except IngestSegmentError as exception:
                logger.error(exception)
                archive_segments.labels(state="pending").inc()

        self.gather_unmanaged_segments()
        self.ingest_pending_segments()
        logger.info("%s stopped", self.name)



import logging
from datetime import datetime
from decimal import Decimal
from pathlib import Path
from queue import Empty, Queue
from subprocess import PIPE, CalledProcessError, run
from threading import Event as ThreadEvent, Thread
from typing import Optional, Union

from pydantic import BaseModel, Field
from typing_extensions import Literal, Protocol, TypeAlias

from .prometheus import stream_silence, stream_status

logger = logging.getLogger(__name__)


def now():
    return datetime.now()


class Event(BaseModel):
    # pylint: disable=unnecessary-lambda
    when: datetime = Field(default_factory=lambda: now())


class SilenceEvent(Event):
    name: Literal["silence"] = "silence"
    kind: Literal["start", "end"]
    seconds: Decimal
    duration: Optional[Decimal] = None


class StatusEvent(Event):
    name: Literal["status"] = "status"
    kind: Literal["up", "down"]


AnyEvent: TypeAlias = Union[SilenceEvent, StatusEvent]


class Hook(Protocol):  # pylint: disable=too-few-public-methods
    def __call__(self, event: AnyEvent):
        pass


class HookError(Exception):
    """
    An error occurred during the hook execution.
    """


class FileHook:  # pylint: disable=too-few-public-methods
    filepath: Path
    log_stderr: bool

    def __init__(self, filepath: str, log_stderr: bool = False) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.is_file():
            raise ValueError(f"hook '{self.filepath}' is not a file!")
        self.log_stderr = log_stderr

    def __call__(self, event: AnyEvent):
        try:
            env = {}
            for key, value in event.model_dump().items():
                if value is not None:
                    if isinstance(value, datetime):
                        value = value.isoformat()
                    env[f"EVENT_{key.upper()}"] = str(value)

            cmd = run(str(self.filepath), env=env, check=True, stderr=PIPE, text=True)
            if self.log_stderr and cmd.stderr:
                logger.info(cmd.stderr)
        except CalledProcessError as exception:
            raise HookError(exception.stderr) from exception


class PrometheusHook:  # pylint: disable=too-few-public-methods
    def __call__(self, event: AnyEvent):
        if isinstance(event, StatusEvent):
            stream_status.state(event.kind)  # pylint: disable=no-member
        elif isinstance(event, SilenceEvent):
            state_map = {"start": "up", "end": "down"}
            stream_silence.state(state_map[event.kind])  # pylint: disable=no-member


class EventHandler(Thread):
    name = "event_handler"
    queue: Queue
    stop: ThreadEvent
    hooks: list[Hook] = []

    def __init__(
        self,
        queue: Queue,
        stop: ThreadEvent,
    ):
        Thread.__init__(self)
        self.queue = queue
        self.stop = stop

    def run(self):
        logger.info("starting %s", self.name)

        while not self.stop.is_set() or not self.queue.empty():
            try:
                event = self.queue.get(timeout=2)
                logger.debug(event)
                for hook in self.hooks:
                    try:
                        hook(event)
                    except HookError as exception:
                        logger.exception(exception)

                self.queue.task_done()
            except Empty:
                pass

        logger.info("%s stopped", self.name)



import logging
from abc import ABC, abstractmethod
from decimal import Decimal
from os import getenv
from queue import Queue
from subprocess import DEVNULL, PIPE, Popen
from threading import Event as ThreadEvent, Thread
from time import sleep
from typing import Optional

import httpx

from .event import SilenceEvent, StatusEvent

logger = logging.getLogger(__name__)

FFMPEG = getenv("FFMPEG_PATH", "ffmpeg")


class StreamListenerHandler(ABC, Thread):
    name: str
    queue: Queue

    @abstractmethod
    def ffmpeg_output(self) -> list[str]:
        pass


# pylint: disable=too-few-public-methods
class StreamListener:
    name = "stream_listener"
    stop: ThreadEvent
    event_queue: Queue
    stream_url: str

    _handlers: list[StreamListenerHandler] = []
    _client: httpx.Client
    _process: Optional[Popen]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        stop: ThreadEvent,
        event_queue: Queue,
        stream_url: str,
        handlers: list[StreamListenerHandler],
    ):
        self.stop = stop
        self.event_queue = event_queue
        self.stream_url = stream_url

        if len(handlers) == 0:
            raise ValueError("stream listener requires at least one handler")

        self._handlers = handlers
        self._client = httpx.Client()
        self._process = None

        for handler in self._handlers:
            logger.debug("starting %s", handler.name)
            handler.start()

    def _ffmpeg_command(self):
        command = [
            *(FFMPEG, "-hide_banner", "-nostats"),
            *("-loglevel", "40"),  # Verbose
            *("-re", "-vn", "-i", self.stream_url),
        ]

        for handler in self._handlers:
            command.extend(["-map", "0"] + handler.ffmpeg_output())

        return command

    def check_stream(self):
        """
        Check the stream until it is available.
        """
        while not self.stop.is_set():
            try:
                with self._client.stream("GET", self.stream_url) as response:
                    response.raise_for_status()
                    self.event_queue.put(StatusEvent(kind="up"))
                    self.event_queue.put(
                        SilenceEvent(kind="end", seconds=Decimal("0.0"))
                    )
                    return

            except (
                httpx.HTTPStatusError,
                httpx.NetworkError,
                httpx.TimeoutException,
                httpx.TransportError,
            ) as error:
                logger.error(f"could not stream from '{self.stream_url}'")
                logger.debug(error)
                self.event_queue.put(StatusEvent(kind="down"))
                sleep(5)

    def run_forever(self):
        while not self.stop.is_set():
            self.check_stream()
            self.listen()

        self._client.close()

        for handler in self._handlers:
            logger.debug("joining %s", handler.name)
            handler.join(timeout=3.0)

    def listen(self):
        logger.info("starting %s", self.name)

        command = self._ffmpeg_command()
        logger.debug(f"running ffmpeg command '{' '.join(command)}'")

        with Popen(
            command,
            bufsize=1,
            stdout=DEVNULL,
            stderr=PIPE,
            text=True,
            encoding="utf-8",
            errors="backslashreplace",
        ) as process:
            self._process = process

            assert self._process.stderr is not None
            for line in iter(self._process.stderr.readline, ""):
                line = line.rstrip()
                logger.debug(line)
                for handler in self._handlers:
                    handler.queue.put(line)

            exit_code = self._process.wait()

        logger.info(f"ffmpeg command exited with {exit_code}")

        logger.info("%s stopped", self.name)

    def stop_listener(self):
        if self._process is not None:
            self._process.terminate()



from prometheus_client import Counter, Enum, Summary

# Internal

stats_scraping = Summary(
    "earhorn_stats_scraping",
    "Time spent extracting Icecast stats",
)

stats_errors = Counter(
    "earhorn_stats_errors",
    "Errors count extracting Icecast stats",
)

# Archive

archive_segments = Counter(
    "earhorn_archive_segments",
    "Archiver segments counter",
    labelnames=["state"],
)

archive_errors = Counter(
    "earhorn_archive_errors",
    "Archiver errors counter",
)

# Stream

stream_status = Enum(
    "earhorn_stream_status",
    "Whether the stream is up",
    states=["up", "down"],
)
stream_silence = Enum(
    "earhorn_stream_silence",
    "Whether the stream is silent",
    states=["up", "down"],
)



import logging
import os
from queue import Queue
from signal import SIGINT, SIGTERM, signal
from threading import Event as ThreadEvent
from typing import Callable, Optional

import click
from dotenv import load_dotenv
from prometheus_client import start_http_server

from . import version
from .event import EventHandler, FileHook, PrometheusHook
from .stats import StatsCollector
from .stream import StreamListener, StreamListenerHandler
from .stream_archive import (
    DEFAULT_ARCHIVE_SEGMENT_FILEPATH,
    DEFAULT_ARCHIVE_SEGMENT_FORMAT,
    DEFAULT_ARCHIVE_SEGMENT_SIZE,
    ArchiveHandler,
    ArchiveStorage,
    LocalArchiveStorage,
)
from .stream_silence import (
    DEFAULT_SILENCE_DETECT_DURATION,
    DEFAULT_SILENCE_DETECT_NOISE,
    DEFAULT_SILENCE_DETECT_RAW,
    SilenceHandler,
)

for module in ("httpx", "urllib3", "botocore", "s3transfer"):
    logging.getLogger(module).setLevel(logging.INFO)
    logging.getLogger(module).propagate = False

logger = logging.getLogger(__name__)


load_dotenv()


# pylint: disable=too-many-arguments,too-many-locals
@click.command(context_settings={"max_content_width": 120})
@click.option(
    "--listen-port",
    envvar="LISTEN_PORT",
    help="Listen port for the prometheus metrics endpoint.",
    default=9950,
    show_default=True,
)
@click.option(
    "--hook",
    envvar="HOOK",
    help="Path to a custom script executed to handle stream state `events`.",
    type=click.Path(),
)
@click.option(
    "--hook-log-stderr",
    envvar="HOOK_LOG_STDERR",
    help="Log the hook stderr output.",
    is_flag=True,
)
@click.option(
    "--stats-url",
    envvar="STATS_URL",
    help="URL to the icecast admin xml stats page.",
)
@click.option(
    "--stats-user",
    envvar="STATS_USER",
    help="Username for the icecast admin xml stats page.",
    default="admin",
    show_default=True,
)
@click.option(
    "--stats-password",
    envvar="STATS_PASSWORD",
    help="Password for the icecast admin xml stats page.",
)
@click.option(
    "--stream-url",
    envvar="STREAM_URL",
    help="URL to the icecast stream.",
)
@click.option(
    "--silence-detect-noise",
    envvar="SILENCE_DETECT_NOISE",
    help="Silence detect noise.",
    default=DEFAULT_SILENCE_DETECT_NOISE,
    show_default=True,
)
@click.option(
    "--silence-detect-duration",
    envvar="SILENCE_DETECT_DURATION",
    help="Silence detect duration.",
    default=DEFAULT_SILENCE_DETECT_DURATION,
    show_default=True,
)
@click.option(
    "--silence-detect-raw",
    envvar="SILENCE_DETECT_RAW",
    help=(
        "Silence detect ffmpeg raw filter. Can be used "
        "to add extra filters (e.g. highpass) or extra "
        "options to the silencedetect filter."
    ),
    default=DEFAULT_SILENCE_DETECT_RAW,
    show_default=True,
)
@click.option(
    "--archive-path",
    envvar="ARCHIVE_PATH",
    help=(
        "Path or url to the archive storage, supported "
        "storage are local filesystem and s3. If defined, "
        "the stream will be archived in the storage as "
        "segments."
    ),
    type=click.Path(),
)
@click.option(
    "--archive-segment-filepath",
    envvar="ARCHIVE_SEGMENT_FILEPATH",
    help="Archive segment filepath.",
    default=DEFAULT_ARCHIVE_SEGMENT_FILEPATH,
    show_default=True,
)
@click.option(
    "--archive-segment-size",
    envvar="ARCHIVE_SEGMENT_SIZE",
    help="Archive segment size in seconds.",
    default=DEFAULT_ARCHIVE_SEGMENT_SIZE,
    show_default=True,
)
@click.option(
    "--archive-segment-format",
    envvar="ARCHIVE_SEGMENT_FORMAT",
    help="Archive segment format.",
    default=DEFAULT_ARCHIVE_SEGMENT_FORMAT,
    show_default=True,
)
@click.option(
    "--archive-segment-format-options",
    envvar="ARCHIVE_SEGMENT_FORMAT_OPTIONS",
    help="Archive segment format options.",
)
@click.option(
    "--archive-copy-stream",
    envvar="ARCHIVE_COPY_STREAM",
    help=(
        "Copy the `stream` without transcoding (reduce CPU usage). "
        "WARNING: The stream has to be in the same format as the "
        "`--archive-segment-format`."
    ),
    is_flag=True,
)
@click.option(
    "--log-level",
    envvar="LOG_LEVEL",
    help="Logging level.",
    default="info",
    show_default=True,
    type=click.Choice(
        ["critical", "error", "warning", "info", "debug"],
        case_sensitive=False,
    ),
)
def cli(
    listen_port: int,
    hook: Optional[str],
    hook_log_stderr: bool,
    stats_url: Optional[str],
    stats_user: str,
    stats_password: str,
    stream_url: Optional[str],
    silence_detect_noise: str,
    silence_detect_duration: str,
    silence_detect_raw: str,
    archive_path: Optional[str],
    archive_segment_filepath: str,
    archive_segment_size: int,
    archive_segment_format: str,
    archive_segment_format_options: Optional[str],
    archive_copy_stream: bool,
    log_level: str,
):
    # pylint: disable=line-too-long
    """
    ENVIRONMENT VARIABLES:

    If a `.env` file is present in the current directory, it will be loaded
    and can be used to pass environment variables to this tool.

    ARCHIVE STORAGE:

    The storage can be defined using a path to a local directory or an
    url to an s3 bucket. Segments will be saved on the storage you
    specified.

    To use an s3 bucket, you need to install the `s3` extras (`pip install earhorn[s3]`),
    use `s3://bucket-name` as value for the `--archive-path` option and export the
    s3 bucket credentials listed in the table below:

    \b
    | Variable                | Description                               | Example                     |
    | ----------------------- | ----------------------------------------- | --------------------------- |
    | AWS_ACCESS_KEY_ID       | The access key for your bucket user       | AKIA568knmklmk              |
    | AWS_SECRET_ACCESS_KEY   | The secret key for your bucket user       | mi0y84wu498zxsasa           |
    | AWS_S3_ENDPOINT_URL     | The endpoint to your s3 bucket (optional) | https://s3.nl-ams.scw.cloud |
    | AWS_S3_REGION_NAME      | Region of your s3 bucket                  | us-east-2                   |

    Example: export AWS_S3_ENPOINT_URL="https://s3.nl-ams.scw.cloud"

    ARCHIVE SEGMENTS:

    \b
    To change the segments duration or format, see the ffmpeg documentation for details
    about the available options:
    https://ffmpeg.org/ffmpeg-formats.html#segment_002c-stream_005fsegment_002c-ssegment

    HOOK SCRIPT:

    To handle events using a custom script, define the `--hook` option with a path to
    your script. This script will not receive any arguments, and the event details will be
    passed via environment variables.

    \b
    | Event   | Environment variables | Values                                                |
    | ------- | --------------------- | ----------------------------------------------------- |
    | status  |                       |                                                       |
    |         | EVENT_NAME            | `status`                                              |
    |         | EVENT_KIND            | `up` or `down`                                        |
    |         | EVENT_WHEN            | ISO 8601 datetime                                     |
    | silence |                       |                                                       |
    |         | EVENT_NAME            | `silence`                                             |
    |         | EVENT_KIND            | `start` or `end`                                      |
    |         | EVENT_WHEN            | ISO 8601 datetime                                     |
    |         | EVENT_DURATION        | Duration in seconds (only available on silence end) |
    """
    logging.basicConfig(
        level=log_level.upper(),
        format="%(asctime)s | %(levelname)-8s | %(name)s:%(funcName)s:%(lineno)s - %(message)s",
    )

    if "SENTRY_DSN" in os.environ:
        logger.info("installing sentry")
        # pylint: disable=import-outside-toplevel
        import sentry_sdk

        sentry_sdk.init(
            traces_sample_rate=1.0,
            release=version,
        )

    if stream_url is None and stats_url is None:
        raise click.UsageError("Specify at least one of --stream-url or --stats-url.")

    # Setup stop mechanism
    stop_event = ThreadEvent()
    stop_functions: list[Callable] = []

    def stop_handler(_signum, _frame):
        logger.info("stopping...")
        stop_event.set()

        for function in stop_functions:
            function()

    signal(SIGINT, stop_handler)
    signal(SIGTERM, stop_handler)

    # Setup event handler before doing any checks
    event_queue: Queue = Queue()
    event_handler = EventHandler(event_queue, stop_event)

    # Setup prometheus hook
    start_http_server(listen_port)
    event_handler.hooks.append(PrometheusHook())

    # Setup file hook
    if hook is not None:
        event_handler.hooks.append(FileHook(hook, hook_log_stderr))

    event_handler.start()

    # Starting stats collector
    if stats_url is not None:
        stats_collector = StatsCollector(
            url=stats_url,
            auth=(stats_user, stats_password),
        )

    # Starting stream listener
    if stream_url is not None:
        handlers: list[StreamListenerHandler] = [
            SilenceHandler(
                stop=stop_event,
                event_queue=event_queue,
                noise=silence_detect_noise,
                duration=silence_detect_duration,
                raw=silence_detect_raw,
            )
        ]

        if archive_path is not None:
            archive_storage: ArchiveStorage

            if archive_path.startswith("s3://"):
                # pylint: disable=import-outside-toplevel
                from .stream_archive_s3 import S3ArchiveStorage

                archive_storage = S3ArchiveStorage(archive_path)
            else:
                archive_storage = LocalArchiveStorage(archive_path)

            handlers.append(
                ArchiveHandler(
                    stop=stop_event,
                    storage=archive_storage,
                    segment_size=archive_segment_size,
                    segment_filepath=archive_segment_filepath,
                    segment_format=archive_segment_format,
                    segment_format_options=archive_segment_format_options,
                    copy_stream=archive_copy_stream,
                )
            )

        stream_listener = StreamListener(
            stop=stop_event,
            event_queue=event_queue,
            stream_url=stream_url,
            handlers=handlers,
        )
        stop_functions.append(stream_listener.stop_listener)
        stream_listener.run_forever()

    event_handler.join()

    if stats_url is not None:
        stats_collector.close()


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



import logging

import httpx
from lxml import etree
from prometheus_client.core import (
    REGISTRY,
    CollectorRegistry,
    GaugeMetricFamily,
    InfoMetricFamily,
)

from .prometheus import stats_errors, stats_scraping

logger = logging.getLogger(__name__)

ICECAST_SOURCE_INFO_KEYS = {
    "server_name",
    "server_description",
    "server_type",
    "audio_info",
    "user_agent",
    "stream_start_iso8601",
}

ICECAST_INFO_KEYS = {
    "admin",
    "host",
    "location",
    "server_id",
    "server_start_iso8601",
}


def gauge_metric_family_with_labels(
    name: str,
    documentation: str,
    value: str,
    labels: dict[str, str],
):
    metric = GaugeMetricFamily(
        name=name,
        documentation=documentation,
        labels=list(labels.keys()),
    )
    metric.add_metric(
        list(labels.values()),
        int(value),
    )
    return metric


def icecast_metrics_factory():
    return InfoMetricFamily(
        name="icecast",
        documentation="Details usually set in the server config, such as: admin, host, "
        "location, server_id, server_start_iso8601.",
    ), {
        "client_connections": GaugeMetricFamily(
            name="icecast_client_connections",
            documentation="Client connections are basically anything that is not a source "
            "connection. These include listeners (not concurrent, but cumulative), "
            "any admin function accesses, and any static content (file serving) "
            "accesses. This is an accumulating counter.",
        ),
        "clients": GaugeMetricFamily(
            name="icecast_clients",
            documentation="Number of currently active client connections.",
        ),
        "connections": GaugeMetricFamily(
            name="icecast_connections",
            documentation="The total of all inbound TCP connections since start-up. "
            "This is an accumulating counter.",
        ),
        "file_connections": GaugeMetricFamily(
            name="icecast_file_connections",
            documentation="This is an accumulating counter.",
        ),
        "listener_connections": GaugeMetricFamily(
            name="icecast_listener_connections",
            documentation="Number of listener connections to mount points. "
            "This is an accumulating counter.",
        ),
        "listeners": GaugeMetricFamily(
            name="icecast_listeners",
            documentation="Number of currently active listener connections.",
        ),
        "source_client_connections": GaugeMetricFamily(
            name="icecast_source_client_connections",
            documentation="Source client connections are the number of times (cumulative since "
            "start-up, not just currently connected) a source client has connected "
            "to Icecast. This is an accumulating counter.",
        ),
        "source_relay_connections": GaugeMetricFamily(
            name="icecast_source_relay_connections",
            documentation="Number of outbound relay connections to (master) icecast servers. "
            "This is an accumulating counter.",
        ),
        "source_total_connections": GaugeMetricFamily(
            name="icecast_source_total_connections",
            documentation="Both clients and relays. This is an accumulating counter.",
        ),
        "sources": GaugeMetricFamily(
            name="icecast_sources",
            documentation="The total of currently connected sources.",
        ),
        "stats": GaugeMetricFamily(
            name="icecast_stats",
            documentation="The total of currently connected STATS clients.",
        ),
        "stats_connections": GaugeMetricFamily(
            name="icecast_stats_connections",
            documentation="Number of times a stats client has connected to Icecast. "
            "This is an accumulating counter.",
        ),
    }


def icecast_source_metrics_factory():
    return InfoMetricFamily(
        name="icecast_source",
        documentation="Details for the Icecast source, such as: server_name, "
        "server_description, server_type, audio_info, user_agent, "
        "stream_start_iso8601, max_listeners.",
        labels=["mount"],
    ), {
        "listener_peak": GaugeMetricFamily(
            name="icecast_source_listener_peak",
            documentation="Peak concurrent number of listener connections for this mountpoint.",
            labels=["mount"],
        ),
        "listeners": GaugeMetricFamily(
            name="icecast_source_listeners",
            documentation="The number of currently connected listeners.",
            labels=["mount"],
        ),
        "slow_listeners": GaugeMetricFamily(
            name="icecast_source_slow_listeners",
            documentation="Number of slow listeners.",
            labels=["mount"],
        ),
        "total_bytes_read": GaugeMetricFamily(
            name="icecast_source_total_bytes_read",
            documentation="Total number of bytes received from the source client.",
            labels=["mount"],
        ),
        "total_bytes_sent": GaugeMetricFamily(
            name="icecast_source_total_bytes_sent",
            documentation="Total number of bytes sent to all listener connections since "
            "last source connect.",
            labels=["mount"],
        ),
    }


# pylint: disable=too-few-public-methods
class StatsCollector:
    """
    Collect and forward stats defined in
    https://icecast.org/docs/icecast-latest/server-stats.html
    """

    _client: httpx.Client
    url: str
    auth: tuple[str, str]

    def __init__(
        self,
        url: str,
        auth: tuple[str, str],
        registry: CollectorRegistry = REGISTRY,
    ):
        self.url = url
        self.auth = auth
        self._client = httpx.Client()

        if registry:
            logger.info("registering stats collector")
            registry.register(self)  # type: ignore

    def close(self):
        logger.info("closing stats collector")
        self._client.close()

    @stats_scraping.time()
    def collect(self):
        try:
            response = self._client.get(self.url, auth=self.auth)
            response.raise_for_status()

        except (
            httpx.ConnectError,
            httpx.HTTPStatusError,
            httpx.ReadError,
            httpx.ReadTimeout,
        ) as error:
            stats_errors.inc()
            logger.error(error)
            return []

        root = etree.fromstring(  # nosec
            response.content,
            parser=etree.XMLParser(resolve_entities=False),
        )

        icecast_info, icecast_metrics = icecast_metrics_factory()
        icecast_source_info, icecast_source_metrics = icecast_source_metrics_factory()

        metrics = []
        metrics.append(icecast_info)
        metrics.extend(icecast_metrics.values())
        metrics.append(icecast_source_info)
        metrics.extend(icecast_source_metrics.values())

        infos: dict[str, str] = {}
        for child in root.iterchildren():
            if child.tag in ICECAST_INFO_KEYS:
                if child.text is not None:
                    infos[child.tag] = child.text

            elif child.tag in icecast_metrics:
                if child.text is not None:
                    icecast_metrics[child.tag].add_metric(
                        [],
                        int(child.text),
                    )

            elif child.tag == "source":
                source_infos: dict[str, str] = {}
                source_labels = [child.attrib.get("mount")]

                for source_child in child.iterchildren():
                    if source_child.tag in ICECAST_SOURCE_INFO_KEYS:
                        if source_child.text is not None:
                            source_infos[source_child.tag] = source_child.text

                    if source_child.tag in icecast_source_metrics:
                        if source_child.text is not None:
                            icecast_source_metrics[source_child.tag].add_metric(
                                source_labels,
                                int(source_child.text),
                            )

                icecast_source_info.add_metric(source_labels, source_infos)
        icecast_info.add_metric([], infos)

        return metrics
